/*
 * Test casefor Manhattan-IDM
 * --Till
 */

#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/applications-module.h"
#include "ns3/mobility-module.h"
#include "ns3/tools-module.h"
#include "ns3/wifi-module.h"
#include "ns3/internet-module.h"

#include <iostream>
#include <fstream>
#include <sstream>
#include <time.h>

#include <vector>
#include <map>
#include <cassert>

#include "svn.h"

#define pp(name) \
  do {                      \
    printSingleParam (#name, c.name); \
  } while (false)

#define cmda(name) \
  do {                      \
    cmd.AddValue(#name, "see elsewhere", c.name); \
  } while (false)

NS_LOG_COMPONENT_DEFINE ("Main");

using namespace ns3;


PacketSocketAddress pbcSocket;

ofstream tfile;
ofstream simfile;
ofstream statfile;


void SingleRun();

struct results {
  std::vector<double> traveltimes;
  std::vector<double> cltimes;
  std::vector<double> handovertimes;
  std::vector<double> electiontimes;
  std::vector<double> requireddecel;
  int leaderConflict;
  int clConflict;
} r;

struct conf {
  // simulation settings
  int svnrev;
  double simtime;
  double warmuptime;
  uint32_t seed;
  uint32_t run;
  
  // topology
  int blocksV;
  int blocksH;
  double blockSizeV;
  double blockSizeH;
  double sourcelength;
  bool maindir;
  bool noconflict;
  
  // wifi
  bool nlos;
  double ibd;
  bool suburban;
  
  // vehicle settings
  double speed;
  double th;
  double acc;
  double dec;
  double vpm;
  double vpmmain;

  
  std::string tl;
  // VTL settings
  std::string vtlleaderstrategy;
  double vtlleaderprob;
  int pgr;
  double clinterval;
  double vtlminphaselength;
  double vtlmaxphaselength;
  
  // STL settings
  bool stlsync;
} c;

template <typename T>
void printSingleParam (std::string name, T value) {
  simfile << "<param name=\"" << name << "\" value=\"" << value <<"\" />" << std::endl; 
}

////////////
// Trace Sinks
////////////

void
VtlBecomesCL (std::string context, Time t)
{
  if (t.GetSeconds() < c.warmuptime)
    return;
  r.cltimes.push_back(t.GetSeconds());
}

void
VtlBecomesVTLLException (std::string context, Time t, double dist)
{
  if (t.GetSeconds() < c.warmuptime)
    return;
      //TODO
}

void
VtlBecomesVTLLHandover (std::string context, Time t)
{
  if (t.GetSeconds() < c.warmuptime)
    return;
  r.handovertimes.push_back(t.GetSeconds());
}

void
VtlBecomesVTLLElection (std::string context, Time t)
{
  if (t.GetSeconds() < c.warmuptime)
    return;
  r.electiontimes.push_back(t.GetSeconds());
}

void
VtlLeaderConflict (std::string context, Time t)
{
  if (t.GetSeconds() < c.warmuptime)
    return;
  r.leaderConflict++;
}

void
CLConflict (std::string context, Time t)
{
  if (t.GetSeconds() < c.warmuptime)
    return;
  r.clConflict++;
}

void  FinishedTrace(std::string context, Time t, double tt) 
{
  if (t.GetSeconds() < c.warmuptime)
    return;
  r.traveltimes.push_back(tt);
}

void 
RequiredDeceleration(std::string context, Time t, double decel) 
{
  if (decel <= 0)
    return;
    
//  if (decel > 2.01)
//    assert(0);
    
  if (t.GetSeconds() < c.warmuptime)
    return;
  r.requireddecel.push_back(decel);
}

void  VtlTrace(std::string context, Time t, uint32_t id, Vector pos, Vector dir, double speed, VtlApplication::Role r, VtlApplication::ApproachState s) {
  tfile << t.GetSeconds() << ";" << id << ";" << pos << ";"<< dir <<";" << speed << ";" << r << ";" << s << std::endl;
}


void  VtlDebugTrace(std::string context, Time t, uint32_t id, Vector pos, Vector dir, double speed, VtlApplication::Role r, VtlApplication::ApproachState s, std::string msg) {
  tfile << t.GetSeconds() << ";" << id << ";" << pos << ";"<< dir <<";" << speed << ";" << r << ";" << s << ";" << msg << std::endl;
}






void addClient (Ptr<Node> tmpNode, double startTime) 
{
  VtlClientHelper vtlh ("ns3::PacketSocketFactory", Address(pbcSocket));
  vtlh.SetAttribute ("StopTime", TimeValue(Seconds(c.simtime+c.warmuptime)));
  vtlh.SetAttribute ("StartTime", TimeValue(Seconds(startTime+1)));
  vtlh.SetAttribute ("Interval", TimeValue(Seconds((double)1.0/c.pgr)));
  vtlh.SetAttribute ("DisableProtocol", BooleanValue(c.tl != "VTL"));
  vtlh.SetAttribute ("CLInterval", TimeValue(Seconds(c.clinterval)));
  vtlh.SetAttribute ("MinPhaseLength", TimeValue(Seconds(c.vtlminphaselength)));
  vtlh.SetAttribute ("MaxPhaseLength", TimeValue(Seconds(c.vtlmaxphaselength)));
  vtlh.SetAttribute ("LeaderElectionStrategy", StringValue(c.vtlleaderstrategy));
  
  vtlh.Install (tmpNode);
}

void setDefaults (void) {
  c.svnrev = SVNREVISION; // included from svn.h
  c.nlos = 1;
  c.warmuptime = 180;
  c.simtime = 300;
  c.tl = "VTL";
  
  c.blocksV = 3;
  c.blocksH = 3;
  c.blockSizeV = 100;
  c.blockSizeH = 100;
  c.sourcelength = 80;
  c.maindir = 0;
  c.noconflict = 1;
  
  c.ibd = 20;
  c.suburban = 0;
  
  c.pgr = 10;
  
  c.speed = 50;
  c.th = 1.5;
  c.acc = 1.5;
  c.dec = 2.5;
  
  c.vpm = 3;
  c.vpmmain = 6;
  c.vtlleaderprob = 0.99;
  
  c.vtlminphaselength = 10;
  c.vtlmaxphaselength = 40;
  c.stlsync = true;
  c.vtlleaderstrategy = "id";
  c.clinterval = 1.0;
  c.seed = SeedManager::GetSeed();
  c.run = SeedManager::GetRun();
}

void getCommandLineParams (int argc, char *argv[]) {
  CommandLine cmd;
  cmda(nlos);
  cmda(simtime);
  cmda(warmuptime);
  cmda(tl);
  cmda(blocksV);
  cmda(blocksH);
  cmda(blockSizeH);
  cmda(blockSizeV);
  cmda(sourcelength);
  cmda(maindir);
  cmda(noconflict);
  cmda(ibd);
  cmda(suburban);
  cmda(pgr);
  cmda(speed);
  cmda(th);
  cmda(acc);
  cmda(dec);
  cmda(vpm);
  cmda(vpmmain);
  cmda(seed);
  cmda(run);
  cmda(clinterval);
  cmda(stlsync);
  cmda(vtlminphaselength);
  cmda(vtlmaxphaselength);
  cmda(vtlleaderstrategy);
  cmda(vtlleaderprob);
  cmd.Parse (argc, argv);  
}



void printParams() {
  pp(svnrev);
  pp(seed);
  pp(run);
  pp(nlos);
  pp(simtime);
  pp(warmuptime);
  pp(tl);
  pp(blocksV);
  pp(blocksH);
  pp(blockSizeH);
  pp(blockSizeV);
  pp(sourcelength);
  pp(maindir);
  pp(noconflict);
  pp(ibd);
  pp(suburban);
  pp(pgr);
  pp(speed);
  pp(th);
  pp(acc);
  pp(dec);
  pp(vpm);
  pp(vpmmain);
  pp(clinterval);
  pp(stlsync);
  pp(vtlminphaselength);
  pp(vtlmaxphaselength);
  pp(vtlleaderstrategy);
  pp(vtlleaderprob);
}


int main (int argc, char *argv[])
{

  setDefaults();

  getCommandLineParams (argc, argv);

  
  SingleRun();
  
}






void ResetResults () {
  r.traveltimes.clear();
  r.cltimes.clear();
  r.handovertimes.clear();
  r.electiontimes.clear();
  r.leaderConflict=0;
  r.clConflict=0;
  r.requireddecel.clear();
}

void WriteListResults(std::vector<double> rlist, std::string name) {
  std::vector<double>::iterator it;
  double startTime = c.warmuptime;
  int sum = 0;

  statfile << "<" << name << " type=\"list\">" << std::endl;
  for (it = rlist.begin(); it != rlist.end(); ++it) {
    if (*it > startTime + 60) {
      statfile << sum << std::endl;
      sum=0;
      startTime += 60;
    }
    sum++;
  }
  statfile << sum << std::endl;
  statfile << "</" << name << ">" << std::endl;
}

void CollectResults () {
  std::vector<double>::iterator it;
  
  statfile << "<nvehicles>" << r.traveltimes.size() << "</nvehicles>" << std::endl;
  
  statfile << "<ncl>" << r.cltimes.size() << "</ncl>" << std::endl;
  WriteListResults(r.cltimes, "clperminute");

  statfile << "<nelection>" << r.electiontimes.size() << "</nelection>" << std::endl;
  WriteListResults(r.electiontimes, "electionsperminute");
  
  statfile << "<nhandover>" << r.handovertimes.size() << "</nhandover>" << std::endl;
  WriteListResults(r.handovertimes, "handoverperminute");
  
  statfile << "<leaderconflict>" << r.leaderConflict << "</leaderconflict>" << std::endl;
  statfile << "<clconflict>" << r.clConflict << "</clconflict>" << std::endl;
  
  statfile << "<traveltimes type=\"list\">" << std::endl;
  for (it = r.traveltimes.begin(); it != r.traveltimes.end(); ++it) {
    statfile << (*it) << std::endl;
  }
  statfile << "</traveltimes>" << std::endl;

  statfile << "<reqdeceleration type=\"list\">" << std::endl;
  for (it = r.requireddecel.begin(); it != r.requireddecel.end(); ++it) {
    statfile << (*it) << std::endl;
  }
  statfile << "</reqdeceleration>" << std::endl;

}

void CloseFiles () {
  simfile << "</sim>" << std::endl;
  statfile << "</statistics>" << std::endl;
    
  simfile.close();
  tfile.close();
  statfile.close();
}

inline bool fileExists (const std::string& name) {
    if (FILE *file = fopen(name.c_str(), "r")) {
        fclose(file);
        usleep(10000);
        return true;
    } else {
        return false;
    }   
}

void OpenFiles() {
  
  std::stringstream filename;
  time_t rawtime;
  struct tm * timeinfo;
  char simfilename [80];
  char tfilename [80];
  char statfilename [80];
  char timestamp [80];
  
  do {
    time (&rawtime);
    timeinfo = localtime (&rawtime);

    strftime (simfilename,80,"%F_%H%M%S_siminfo.xml",timeinfo);
    strftime (tfilename,80,"%F_%H%M%S.tfile",timeinfo);
    strftime (statfilename,80,"%F_%H%M%S_statistics.xml",timeinfo);
    strftime (timestamp,80,"%F_%T",timeinfo);
  } while (fileExists(simfilename));
  
  simfile.open (simfilename, ios::out);
  tfile.open (tfilename, ios::out);
  statfile.open (statfilename, ios::out);
  simfile.precision(13);
  
  simfile << "<sim>" << std::endl;
  statfile << "<statistics>" << std::endl;
  simfile << "<tfile>" << tfilename << "</tfile>" << std::endl;
  simfile << "<statfile>" << statfilename << "</statfile>" << std::endl;
  simfile << "<timestamp>" << timestamp << "</timestamp>" << std::endl;
  
  
}

void SingleRun() {
  


  ResetResults();
  OpenFiles ();
  printParams ();
  
  
  Ptr<VtlCalc> vc = VtlCalc::GetPtr(); 
  vc->p.prob = c.vtlleaderprob;
  
  
  // disable fragmentation
  Config::SetDefault ("ns3::WifiRemoteStationManager::FragmentationThreshold", StringValue ("2200"));
  Config::SetDefault ("ns3::WifiRemoteStationManager::RtsCtsThreshold", StringValue ("2200"));

  
  //LogComponentEnable("YansWifiPhy", LOG_LEVEL_DEBUG);
  //LogComponentEnable("TrafficLight", LOG_LEVEL_DEBUG);
  //LogComponentEnable("VtlApplication", LOG_LEVEL_DEBUG);
  //LogComponentEnable("ManhattanIDM", LOG_LEVEL_DEBUG);
  //LogComponentEnable("IntelligentDriverMobilityModel", LOG_LEVEL_DEBUG);
  //LogComponentEnable("VirtualSource11pPropagationLossModel", LOG_LEVEL_DEBUG);
  //LogComponentEnableAll (LOG_LEVEL_DEBUG);  böse!


  SeedManager::SetSeed(c.seed);
  SeedManager::SetRun(c.run);



  // Create a socket for the to be attached application

  pbcSocket.SetAllDevices();
  pbcSocket.SetPhysicalAddress(Mac48Address::GetBroadcast());
  pbcSocket.SetProtocol(1);
  
  
  ManhattanTopologyHelper mth;
  
  mth.speedvar = NormalVariable  (c.speed/3.6, (3/3.6)*(3/3.6), 9/3.6);
  mth.thvar = NormalVariable (c.th, 0.2*0.2, 0.5);
  mth.accvar = NormalVariable (c.acc, 0.2*0.2, 0.5);
  mth.decvar = NormalVariable (c.dec, 0.2*0.2, 0.5);  
  mth.ivtvar = ExponentialVariable (60.0/c.vpm);
  mth.ivtmainvar = ExponentialVariable (60.0/c.vpmmain);
  mth.m_sourceLength = c.sourcelength;
  mth.m_makeTrafficLights = (c.tl == "STL");
  mth.m_stlsync = c.stlsync;
  mth.m_callback = MakeCallback(&addClient);
  

  mth.m_mainDir = c.maindir;
  mth.m_setNoConflict = c.noconflict;
  mth.makeSimpleTopo (c.blocksV, c.blocksH, c.blockSizeV, c.blockSizeH, c.ibd, c.suburban, Seconds(c.simtime+c.warmuptime));


  simfile << mth.getIntersectionLayout() << std::endl;


  
  // The below set of helpers will help us to put together the wifi NICs we want
  WifiHelper wifi;
  //wifi.SetStandard (WIFI_PHY_STANDARD_80211_10Mhz);

  YansWifiPhyHelper wifiPhy =  YansWifiPhyHelper::Default ();

  /*
   * see http://www.nsnam.org/doxygen-release/classns3_1_1_yans_wifi_phy.html
   * for explaination of parameters
   */
  wifiPhy.Set ("RxGain", DoubleValue (1.5) ); 
  wifiPhy.Set ("TxGain", DoubleValue (1.5) ); 
  wifiPhy.Set ("TxPowerEnd", DoubleValue (20) );
  wifiPhy.Set ("TxPowerStart", DoubleValue (20) );
  wifiPhy.Set ("TxPowerLevels", UintegerValue (1) );
  wifiPhy.Set ("EnergyDetectionThreshold", DoubleValue (-92) ); //default: -96
  wifiPhy.Set ("CcaMode1Threshold", DoubleValue (-92) );
  wifiPhy.Set ("RxNoiseFigure", DoubleValue (1) );
  
  
    // used for pcap tracing (tcpdump)
  wifiPhy.SetPcapDataLinkType (YansWifiPhyHelper::DLT_IEEE802_11_RADIO); 

  /*
   * now we add propagation models to our channel.
   * see http://www.nsnam.org/doxygen/classns3_1_1_propagation_loss_model.html
   * for list of available models (models can be chained!)
   */
  YansWifiChannelHelper wifiChannel ;
  wifiChannel.SetPropagationDelay ("ns3::ConstantSpeedPropagationDelayModel");
  
  
  if (c.nlos && (c.tl != "STL")) {
    // add NLOS propagation loss model. systemloss is set w/o antenna gain as
    // antenna gain is set in wifiPhy
    wifiChannel.AddPropagationLoss ("ns3::VirtualSource11pPropagationLossModel",
                                          "HeightAboveZ", DoubleValue (1.5),
                                          "SystemLoss", DoubleValue (4.8),
                                          "RoadSize", DoubleValue (10.5),
                                          "IsSuburban", BooleanValue (c.suburban),
                                          "CrossingCenter", VectorValue (Vector (0, 0, 0)));

  } else {
    // perfect channel
    wifiChannel.AddPropagationLoss ("ns3::RangePropagationLossModel",
                                            "MaxRange", DoubleValue (2000));
  }
  
  wifiPhy.SetChannel (wifiChannel.Create ());
  
  // Add a non-QoS upper mac, and disable rate control
  NqosWifiMacHelper wifiMac = NqosWifiMacHelper::Default ();
  
  std::string wifiMode = "OfdmRate6MbpsBW10MHz";
  //std::string wifiMode = "OfdmRate6Mbps";
  
  wifi.SetRemoteStationManager("ns3::ConstantRateWifiManager",
                                 "DataMode", StringValue(wifiMode),
                                 "NonUnicastMode", StringValue(wifiMode));

  // Set it to adhoc mode
  wifiMac.SetType ("ns3::AdhocWifiMac");
  NetDeviceContainer devices = wifi.Install (wifiPhy, wifiMac, mth.m_nc);
  //wifiPhy.EnablePcap ("nlos", devices);


  
  
///////////////////// SOCKET

  PacketSocketHelper packetSocket;
  packetSocket.Install (mth.m_nc);
  


  Config::Connect("/NodeList/*/ApplicationList/*/$ns3::VtlApplication/BecomesCL", MakeCallback(&VtlBecomesCL));
  Config::Connect("/NodeList/*/ApplicationList/*/$ns3::VtlApplication/BecomesVTLLbyException", MakeCallback(&VtlBecomesVTLLException));
  Config::Connect("/NodeList/*/ApplicationList/*/$ns3::VtlApplication/BecomesVTLLbyHandover", MakeCallback(&VtlBecomesVTLLHandover));
  Config::Connect("/NodeList/*/ApplicationList/*/$ns3::VtlApplication/BecomesVTLLbyElection", MakeCallback(&VtlBecomesVTLLElection));
  Config::Connect("/NodeList/*/ApplicationList/*/$ns3::VtlApplication/CLConflict", MakeCallback(&CLConflict));
  Config::Connect("/NodeList/*/ApplicationList/*/$ns3::VtlApplication/LeaderConflict", MakeCallback(&VtlLeaderConflict));
  Config::Connect("/NodeList/*/ApplicationList/*/$ns3::VtlApplication/Finished", MakeCallback(&FinishedTrace));
  Config::Connect("/NodeList/*/ApplicationList/*/$ns3::VtlApplication/RequiredDeceleration", MakeCallback(&RequiredDeceleration));
    
  Config::Connect("/NodeList/*/ApplicationList/*/$ns3::VtlApplication/VtlUpdate", MakeCallback(&VtlTrace));
  Config::Connect("/NodeList/*/ApplicationList/*/$ns3::VtlApplication/VtlDebugTrace", MakeCallback(&VtlDebugTrace));

  
  
  Simulator::Run ();

  Simulator::Destroy ();
  
  CollectResults ();
  
  CloseFiles ();
  std::cout << "Finished run" << std::endl;
}
