/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2008 INRIA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Mohamed Amine Ismail <amine.ismail@sophia.inria.fr>
 */
#include "pbc-helper.h"
#include "ns3/pbc.h"
#include "ns3/uinteger.h"
#include "ns3/string.h"

namespace ns3 {



PbcClientHelper::PbcClientHelper (std::string protocol, Address address)
{
  m_factory.SetTypeId (PbcApplication::GetTypeId ());
  m_factory.Set ("Protocol", StringValue (protocol));
  m_factory.Set ("RemoteAddress", AddressValue (address));
}

PbcClientHelper::PbcClientHelper ()
{
}


void
PbcClientHelper::SetAttribute (std::string name, const AttributeValue &value)
{
  m_factory.Set (name, value);
}

ApplicationContainer
PbcClientHelper::Install (NodeContainer c)
{
  ApplicationContainer apps;
  for (NodeContainer::Iterator i = c.Begin (); i != c.End (); ++i)
    {
      Ptr<Node> node = *i;
      Ptr<PbcApplication> client = m_factory.Create<PbcApplication> ();
      node->AddApplication (client);
      apps.Add (client);
    }
  return apps;
}

ApplicationContainer
PbcClientHelper::Install (Ptr<Node> node)
{
  ApplicationContainer apps;
  Ptr<PbcApplication> client = m_factory.Create<PbcApplication> ();
  node->AddApplication (client);
  apps.Add (client);
  return apps;
}


} // namespace ns3
