/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
//
//


#include "ns3/log.h"
#include "ns3/address.h"
#include "ns3/node.h"
#include "ns3/nstime.h"
#include "ns3/data-rate.h"
#include "ns3/random-variable.h"
#include "ns3/socket.h"
#include "ns3/simulator.h"
#include "ns3/socket-factory.h"
#include "ns3/packet.h"
#include "ns3/tag.h"
#include "ns3/uinteger.h"
#include "ns3/integer.h"
#include "ns3/string.h"
#include "ns3/boolean.h"
#include "ns3/trace-source-accessor.h"
#include "vtl-client.h"
#include "ns3/udp-socket-factory.h"
#include "ns3/inet-socket-address.h"
#include "ns3/double.h"
#include "ns3/node-list.h"

#include <math.h>
#include "ns3/virtualsource11p-propagation-loss-model.h"

NS_LOG_COMPONENT_DEFINE ("VtlApplication");

using namespace std;

namespace ns3 {


NS_OBJECT_ENSURE_REGISTERED (VtlHeader);

TypeId
VtlHeader::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::VtlHeader")
    .SetParent<Header> ()
    .AddConstructor<VtlHeader> ()
    ;
  return tid;
}

VtlHeader::VtlHeader ()
  : m_source (0),
    m_seqNo (0),
    m_speed (0),
    m_intersectionId (VTL_INT_INVALID),
    m_localId (VTL_BYTE_INVALID),
    m_isClusterLeader (false),
    m_isLeader (false),
    m_electionReady (false),
    m_nextLeader (VTL_INT_INVALID),
    m_clusterSize (0),
    m_dist (0),
    m_nVtlInfo (0)
{}


VtlHeader::~VtlHeader ()
{}

TypeId
VtlHeader::GetInstanceTypeId (void) const
{
  return GetTypeId ();
}
uint32_t
VtlHeader::GetSerializedSize (void) const
{
  uint32_t sum = 45+45  // static
                 + m_nVtlInfo*(8+8+VTL_HEADER_MAXROADS);
  return sum;
}
void
VtlHeader::Serialize (Buffer::Iterator start) const
{
  Buffer::Iterator i = start;
  i.WriteU32 (m_source);                      // 4
  i.WriteU32 (m_seqNo);                       // 4
  i.WriteU64 (m_timestamp.GetMicroSeconds()); // 8
  i.Write ((uint8_t*)&m_pos.x, 8);            // 8
  i.Write ((uint8_t*)&m_pos.y, 8);            // 8
  i.Write ((uint8_t*)&m_speed, 8);            // 8

  i.WriteU32 (m_intersectionId);              // 4
  i.WriteU8 (m_localId);                      // 1  sum = 45
  
  
  i.WriteU8 (m_isClusterLeader);                    // 1
  i.WriteU64 (m_clValidity.GetMicroSeconds());      // 8
  i.WriteU8 (m_isLeader);                           // 1
  i.WriteU64 (m_leaderValidity.GetMicroSeconds());  // 8
  i.WriteU8 (m_electionReady);                      // 1
  i.WriteU64 (m_erValidity.GetMicroSeconds());      // 8 
  
  i.WriteU32 (m_nextLeader);                        // 4
  i.WriteU32 (m_currentPhase);                      // 4
  i.WriteU8 (m_clusterSize);                        // 1
  i.Write ((uint8_t*)&m_dist, 8);                   // 8
  i.WriteU8 (m_nVtlInfo);                           // 1 sum = 45


  // vtl info
  // size = m_nVtlInfo*(8+8+VTL_HEADER_MAXROADS)
  int j;
  for (j = 0; j < m_nVtlInfo; ++j) {
    i.WriteU64 (m_vtlInfo[j].validFrom.GetMicroSeconds());
    i.WriteU64 (m_vtlInfo[j].validUntil.GetMicroSeconds());
    i.Write (m_vtlInfo[j].lightmap, VTL_HEADER_MAXROADS); 
  }
  
  
}


uint32_t
VtlHeader::Deserialize (Buffer::Iterator start)
{
  
  m_source = start.ReadU32 ();                        // 4
  m_seqNo = start.ReadU32 ();                         // 4
  m_timestamp = MicroSeconds (start.ReadU64());       // 8
  start.Read ((uint8_t*)&m_pos.x, 8);                 // 8
  start.Read ((uint8_t*)&m_pos.y, 8);                 // 8
  start.Read ((uint8_t*)&m_speed, 8);                 // 8
  
  m_intersectionId = start.ReadU32 ();                // 4
  m_localId = start.ReadU8 ();                        // 1  sum = 45
  
  m_isClusterLeader = start.ReadU8 ();                // 1
  m_clValidity = MicroSeconds (start.ReadU64());      // 8
  m_isLeader = start.ReadU8 ();                       // 1
  m_leaderValidity = MicroSeconds (start.ReadU64());  // 8
  m_electionReady = start.ReadU8 ();                  // 1
  m_erValidity = MicroSeconds (start.ReadU64());      // 8
  
  m_nextLeader = start.ReadU32 ();                    // 4
  m_currentPhase = start.ReadU32 ();                  // 4
  m_clusterSize = start.ReadU8 ();                    // 1
  start.Read ((uint8_t*)&m_dist, 8);                  // 8
  m_nVtlInfo = start.ReadU8 ();                       // 1 sum = 45

  int j;
  for (j = 0; j < m_nVtlInfo; ++j) {
    m_vtlInfo[j].validFrom = MicroSeconds (start.ReadU64());
    m_vtlInfo[j].validUntil = MicroSeconds (start.ReadU64());
    start.Read (m_vtlInfo[j].lightmap, VTL_HEADER_MAXROADS); 
  }
                                                      // m_nVtlInfo * (8+8+VTL_HEADER_MAXROADS)

  return 45 + 45 + m_nVtlInfo * (8+8+VTL_HEADER_MAXROADS);
}
void
VtlHeader::Print (std::ostream &os) const
{
  /*os << "source=" << m_source << ", seqNo=" << m_seqNo << ", timestamp=" << m_timestamp
    << " position=" << m_pos << " speed=" << m_speed << " dirID=" << (int)m_dirId
    << " VTLLeader=" << m_isLeader << " nextLeader=" << m_nextLeader << " clusterLeader=" << m_isClusterLeader
    << " clusterSize=" << (int)m_clusterSize << " dist=" << m_dist << " greenRoads=" << (int)m_greenRoads
    << " nextChange=" << m_nextChange.GetSeconds() << " state=" << m_state;
   */
}







NS_OBJECT_ENSURE_REGISTERED (VtlApplication);

TypeId
VtlApplication::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::VtlApplication")
    .SetParent<Application> ()
    .AddConstructor<VtlApplication> ()
    .AddAttribute ("Interval",
                   "The time interval between two successive broadcast transmissions.",
                   TimeValue (Seconds(0.1)),
                   MakeTimeAccessor (&VtlApplication::m_interval),
                   MakeTimeChecker())
    .AddAttribute ("CLInterval",
                   "Time to wait before becoming CL.",
                   TimeValue (Seconds(1.0)),
                   MakeTimeAccessor (&VtlApplication::m_clusterInterval),
                   MakeTimeChecker())
    .AddAttribute ("DisableProtocol",
                   "Do not run VTL protocol but just let mobility drive.",
                   BooleanValue (false),
                   MakeBooleanAccessor (&VtlApplication::m_disabled),
                   MakeBooleanChecker())
    .AddAttribute ("MinPhaseLength",
                   "Minimum length of traffic light phase",
                   TimeValue (Seconds(10)),
                   MakeTimeAccessor (&VtlApplication::m_minPhaseLength),
                   MakeTimeChecker())
    .AddAttribute ("MaxPhaseLength",
                   "Minimum length of traffic light phase",
                   TimeValue (Seconds(40)),
                   MakeTimeAccessor (&VtlApplication::m_maxPhaseLength),
                   MakeTimeChecker())
    .AddAttribute ("LeaderTimeslicePretime",
                   "The time before the last VTL information expires the VTL leader sets the next VTL info",
                   TimeValue (Seconds(3)),
                   MakeTimeAccessor (&VtlApplication::m_leaderTimeslicePretime),
                   MakeTimeChecker())
    .AddAttribute ("PacketSize",
                   "Packet size in Bytes.",
                   IntegerValue (400),
                   MakeIntegerAccessor (&VtlApplication::m_packetSize),
                   MakeIntegerChecker<uint32_t> ())
    .AddAttribute ("MaxAcc",
                   "Maximum assumed acceleration.",
                   DoubleValue (2.0),
                   MakeDoubleAccessor (&VtlApplication::m_maxAcc),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("SafeAcc",
                   "Safe deceleration.",
                   DoubleValue (2.0),
                   MakeDoubleAccessor (&VtlApplication::m_safeAcc),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("ReactionTime",
                   "Reaction Time in seconds",
                   DoubleValue (0.0),
                   MakeDoubleAccessor (&VtlApplication::m_reactionTime),
                   MakeDoubleChecker<double> ())
    .AddAttribute ("RemoteAddress",
                   "The address of the destination (either a Unicast or a Broadcast address)",
                   AddressValue (),
                   MakeAddressAccessor (&VtlApplication::m_peer),
                   MakeAddressChecker ())
    .AddAttribute ("Protocol",
                   "The type of protocol to use.",
                   TypeIdValue (UdpSocketFactory::GetTypeId ()),
                   MakeTypeIdAccessor (&VtlApplication::m_protocol),
                   MakeTypeIdChecker ())
    .AddAttribute ("LeaderElectionStrategy",
                   "How to decide if to become leader",
                   StringValue ("id"),
                   MakeStringAccessor (&VtlApplication::m_leaderStrategy),
                   MakeStringChecker ())
    .AddTraceSource ("BecomesCL",
                   "Vehicle becomes cluster leader",
                   MakeTraceSourceAccessor (&VtlApplication::m_clTrace))
    .AddTraceSource ("RequiredDeceleration",
                   "Required deceleration on braking",
                   MakeTraceSourceAccessor (&VtlApplication::m_requiredDecelerationTrace))
    .AddTraceSource ("BecomesVTLLbyException",
                   "Vehicle becomes VTL leader after leader election threshold",
                   MakeTraceSourceAccessor (&VtlApplication::m_vtllExceptionTrace))
    .AddTraceSource ("BecomesVTLLbyHandover",
                   "Vehicle becomes cluster leader by handover",
                   MakeTraceSourceAccessor (&VtlApplication::m_vtllHandoverTrace))
    .AddTraceSource ("BecomesVTLLbyElection",
                   "Vehicle becomes cluster leader by election",
                   MakeTraceSourceAccessor (&VtlApplication::m_vtllElectionTrace))
    .AddTraceSource ("LeaderConflict",
                   "A vehicle detects a leader conflict",
                   MakeTraceSourceAccessor (&VtlApplication::m_leaderConflictTrace))
    .AddTraceSource ("CLConflict",
                   "A vehicle detects a CL conflict",
                   MakeTraceSourceAccessor (&VtlApplication::m_clConflictTrace))
    .AddTraceSource ("Finished",
                   "A vehicle has passed the intersection",
                   MakeTraceSourceAccessor (&VtlApplication::m_finishedTrace))
    .AddTraceSource ("VtlUpdate", 
                   "Regular update about each vehicles position and VTL state",
                   MakeTraceSourceAccessor (&VtlApplication::m_vtlTrace))
    .AddTraceSource ("VtlDebugTrace", 
                   "Traceoutput to a function (e.g. to write to file)",
                   MakeTraceSourceAccessor (&VtlApplication::m_vtlDebugTrace))                    
    ;
  return tid;
}


VtlApplication::VtlApplication()
 : m_socket(0),
   m_seqNo(0),
   m_id (0)
{
  NS_LOG_FUNCTION_NOARGS ();
  
  NS_ASSERT (IDM_INT_INVALID == VTL_INT_INVALID);
  
  m_state = SLAVE;
  
  //memset(&m_vtl, 0, sizeof(m_vtl));
}

VtlApplication::~VtlApplication()
{
  NS_LOG_FUNCTION_NOARGS ();
}

void
VtlApplication::DoDispose (void)
{
  NS_LOG_FUNCTION_NOARGS ();

  m_socket = 0;

    // chain up
  Application::DoDispose ();

}

void
VtlApplication::ResetOtherCL (void)
{
  uint32_t i;
  for (i = 0; i < VTL_HEADER_MAXROADS; ++i) {
    m_clmap[i].id = VTL_INT_INVALID;
  }
  m_clObservationStart = Simulator::Now();
}


/*
 * called by RoadUpdate (road changed)
 * at start of application and at exception situations (VTL conflict)
 */
void
VtlApplication::Reset (void)
{
  SetState (SLAVE);
  m_approach = FAILSAFE;
  m_handoverTo = VTL_INT_INVALID;
  
  m_mob->SetApproach (IntelligentDriverMobilityModel::FAILSAFE);
  m_requiredDecelerationTrace (Simulator::Now(), GetRequiredDeceleration(m_speed, m_dist));
  Simulator::Cancel(m_leaderUpdateEvent);
  Simulator::Cancel(m_leaderHandoverDegradeEvent);
  
  ResetOtherCL ();
  m_vtlList.clear ();
  m_foundBetterClusterLeaderTime = Simulator::Now ();
  m_ownCluster.clear();
}

void
VtlApplication::StartApplication()
{
  NS_LOG_FUNCTION_NOARGS ();
  
  
  if (m_leaderStrategy == "id")
    e_leaderStrat = ID;
  else if (m_leaderStrategy == "distance")
    e_leaderStrat = DISTANCE;
  else if (m_leaderStrategy == "distancesafe")
    e_leaderStrat = DISTANCESAFE;
  else if (m_leaderStrategy == "probabilistic")
    e_leaderStrat = PROBABILISTIC;
  
  // init position and direction
  m_mob = GetNode()->GetObject<IntelligentDriverMobilityModel>();
  m_navi = m_mob->m_navi;
  m_id = GetNode()->GetId(); 
  m_dir = m_navi->GetCurrentDirection ();
  m_intersectionId = m_navi->GetCurrentIntersectionId ();
  m_localId = m_navi->GetCurrentLocalId ();
  m_vtlcalc = VtlCalc::GetPtr(); 
  UpdatePositionAndSpeed();
  
  
  // subscribe to road updates
  m_navi->SetUpdateCallback (MakeCallback (&VtlApplication::UpdateRoad, this));
  

  Reset ();
  
  
    // Create the socket if not already
  if (!m_socket)
    {
      m_socket = Socket::CreateSocket (GetNode(), m_protocol);
      m_socket->SetRecvCallback (MakeCallback (&VtlApplication::ReceivePacket, this));
      m_socket->Bind ();
      m_socket->Connect (m_peer);
      
    }
    
    // Ensure no pending event
  CancelEvents ();


    // Start application with jitter
  jitterGenerator = UniformVariable(0, m_interval.GetSeconds ());
  double jitterInSeconds = jitterGenerator.GetValue();
  Time m_startJitter = Seconds(jitterInSeconds);

  VTL_LOG_DEBUG ("Start on roadID " << m_navi->m_curRoad->m_id << " with jitter = " << jitterInSeconds << " Seconds");
  m_foundBetterClusterLeaderTime = Simulator::Now () + m_startJitter;
  m_timestepEvent = Simulator::Schedule(m_startJitter, &VtlApplication::NextTimestep, this);
  
  jitterGenerator = UniformVariable(0, m_interval.GetSeconds ()/10.0);
}

void
VtlApplication::StopApplication()
{
  NS_LOG_FUNCTION_NOARGS ();
  SetState (FINISHED);
  CancelEvents ();
  if(m_socket != 0)
    {
      m_socket->Close ();
    }
  else
    {
      NS_LOG_WARN("VtlApplication found null socket to close in StopApplication");
    }
}

void
VtlApplication::CancelEvents ()
{
  NS_LOG_FUNCTION_NOARGS ();
  Simulator::Cancel(m_timestepEvent);
  Simulator::Cancel(m_leaderUpdateEvent);
  Simulator::Cancel(m_sendEvent);
  Simulator::Cancel(m_leaderHandoverDegradeEvent);
}

void
VtlApplication::ScheduleNextTx()
{
  NS_LOG_FUNCTION_NOARGS ();
  Time nextTime(m_interval);
  NS_LOG_LOGIC ("VtlApplication::ScheduleNextTx() nextTime = " << nextTime.GetSeconds() << "sec");
  m_timestepEvent = Simulator::Schedule(nextTime, &VtlApplication::NextTimestep, this);
}

void
VtlApplication::SetState (Role s)
{
  VTL_LOG_DEBUG ("State transition from " << GetStateName (m_state) << " to "  << GetStateName (s));
  m_state = s;
}

std::string
VtlApplication::GetStateName (Role s)
{
  switch (s) {
    case SLAVE:
      return "SLAVE";
    case CL:
      return "CL";
    case LEADER:
      return "LEADER";
    case LEADER_HANDOVER:
      return "LEADER_HANDOVER";
    case FINISHED:
      return "FINISHED";
  }
  return "UNKNOWN";
}

void 
VtlApplication::UpdateRoad (uint32_t intersectionId, uint8_t localId) 
{
  VTL_LOG_DEBUG ("Changed road, next intersection = " << intersectionId);

  Reset ();
  
  if (intersectionId == VTL_INT_INVALID) {
    // vehicle finished!
    SetState (FINISHED);
    
  }
  
  m_intersectionId = intersectionId;
  m_localId = localId;
  m_dir = m_navi->GetCurrentDirection ();
  VtlApplication::UpdatePositionAndSpeed ();
  
  Simulator::Cancel(m_sendEvent);
}

void
VtlApplication::ReceivePacket (Ptr<Socket> socket)
{
  
  NS_LOG_FUNCTION (this << socket);
  while (m_socket->GetRxAvailable () > 0)
  {
    Address from;
    Ptr<Packet> p = m_socket->RecvFrom (0xffffffff, 0, from);
    
    VtlHeader head;
    p->PeekHeader (head);

    uint32_t senderId = head.m_source;
    
    Time timestamp = head.m_timestamp;
    uint32_t iid = head.m_intersectionId;
    
    if (m_state == FINISHED)
      return;
    if (iid != m_intersectionId)
      return; // ignore packets from other intersections
    
    
    uint8_t senderLId = head.m_localId;
    
    if (!m_disabled) {
      
      // Clustering must be done in every state
      // Slaves can become CL when old CL leaves the intersection
      // CL can become Slave when new vehicle appears between itself and intersection
      
      // check if vehicles are on same road in same direction
      if (senderLId == m_localId) {
        // vehicles are driving in the same direction
        
        // check if vehicle is in front or behind or if vtl leader is on my road
        if (head.m_dist < m_dist) {
          // other vehicle is between me and intersection
          if (m_foundBetterClusterLeaderTime < head.m_clValidity)
            m_foundBetterClusterLeaderTime = head.m_clValidity;
          if (m_state != SLAVE) {
            m_clConflictTrace(Simulator::Now());
            VTL_LOG_DEBUG ("Found better Cluster Leader (CLConflict): " << senderId 
            << " mydist = " << m_dist
            << " other dist = " << head.m_dist);
            Reset();
          }
        }
      }
      
      
      // store conflicting CL and store vehicles on own road
      if (m_state == CL || m_state == LEADER) {
        if (head.m_isClusterLeader) {
          if (senderLId != m_localId) {
            // CL on conflicting road
            
            // store information on other CL
            StoreInfo(m_clmap[senderLId], head);
          }
        } else {
          if (senderLId == m_localId) { 
            // slave vehicle on same road
            // store in map for timeslice calculation
            VehicleInfo v;
            StoreInfo (v, head);
            m_ownCluster[senderId] = v;
          }
        }
      }

      // watch out for leader handover messages
      if (m_state != LEADER && m_state != LEADER_HANDOVER) {
        if (head.m_nextLeader == m_id) {
          UpdateVtlInfo (head);
          VTL_LOG_DEBUG ("VTL Handover from Vehicle " << senderId);
          m_vtllHandoverTrace(Simulator::Now());
          SetState (LEADER);
          m_currentPhase = head.m_currentPhase;
          m_roadmap = m_navi->m_curInt->GetRoadMap ();
          m_leaderUpdateEvent = Simulator::Schedule(m_vtlList.back().validUntil - 
            m_leaderTimeslicePretime - Simulator::Now (), 
            &VtlApplication::LeaderUpdateTimeslice, this);
        }
      }
      
      // check for VTL Leader conflict
      if (m_state == LEADER) {
        if (head.m_isLeader && head.m_nextLeader != m_id) {
          m_leaderConflictTrace (Simulator::Now());
          VTL_LOG_DEBUG ("VTL Leader conflict detected with Vehicle " << senderId);
          Reset();
        }
      }
    } // (!m_disabled)
    
    // extract VTL information if attached
    if ((m_state == CL || m_state == SLAVE) && head.m_nVtlInfo > 0) {
      UpdateVtlInfo (head);
    }
    
  }    
}

void
VtlApplication::StoreInfo (VehicleInfo &vi, const VtlHeader &h) 
{
  vi.id = h.m_source;
  vi.isCL = h.m_isClusterLeader;
  vi.clValidity = h.m_clValidity;
  vi.isLeader = h.m_isLeader;
  vi.dist = h.m_dist;
  vi.speed = h.m_speed;
  vi.localId = h.m_localId;
  vi.clusterSize = h.m_clusterSize;
  vi.timestamp = h.m_timestamp;
}

/*
 * processes the VTL information of an incoming packet. only new information
 * is appended to m_vtlList. Information in m_vtlList and the vtlInfo array 
 * is assumed to be sorted in chronological order. 
 */
void
VtlApplication::UpdateVtlInfo (const VtlHeader &h) 
{
  uint8_t nEntries = h.m_nVtlInfo;
  uint8_t i;
  Time minTime (0);
  bool added = false;
  bool wasEmpty = m_vtlList.empty ();
  // get last 
  if (m_vtlList.size () != 0) {
    minTime = m_vtlList.back ().validUntil;
  }
  
  for (i = 0; i < nEntries; ++i) {
    // only add entry to own list if validfrom is after last known vtl info time
    if (minTime <= h.m_vtlInfo[i].validFrom  && Simulator::Now () < h.m_vtlInfo[i].validUntil) {
      m_vtlList.push_back (h.m_vtlInfo[i]);
      NS_ASSERT(m_vtlList.size() <= VTL_HEADER_MAXENTRIES);
      added = true;
      Simulator::Schedule(h.m_vtlInfo[i].validUntil - Simulator::Now (), &VtlApplication::VtlExpire, this);
      VTL_LOG_DEBUG ("Received VTL Info for time " 
        << h.m_vtlInfo[i].validFrom.GetSeconds () << " until " 
        << h.m_vtlInfo[i].validUntil.GetSeconds () << " value: " 
        << ((h.m_vtlInfo[i].lightmap[m_localId] == 1) ? "green" : "red") );
    }
  }
  
  if (wasEmpty && added) {
    CheckCurrentVtl ();
  }
}

void
VtlApplication::VtlExpire (void)
{
  if (m_vtlList.empty ())
    return;
  if (m_state == FINISHED)
    return;
    
  if (m_vtlList.front().validUntil <= Simulator::Now ()) {
    m_vtlList.pop_front ();
    CheckCurrentVtl ();
  }
}

void
VtlApplication::CheckCurrentVtl (void)
{
  if (m_state == FINISHED)
    return;
    
  if (m_vtlList.empty ()) {
    m_approach = FAILSAFE;
    m_mob->SetApproach (IntelligentDriverMobilityModel::FAILSAFE);
    m_requiredDecelerationTrace (Simulator::Now(), GetRequiredDeceleration(m_speed, m_dist));
  } else {
    if (Simulator::Now () <= m_vtlList.front().validUntil) {
      // vtl is still valid
      if (m_vtlList.front().validFrom <= Simulator::Now ()) {
        // vtl is already valid
        m_approach = VALIDVTL;
        if (HasGreen()) {
          VTL_LOG_DEBUG ("Green light -> Go!");
          m_mob->SetApproach (IntelligentDriverMobilityModel::GREEN);
          // check for possible handover in case we are leader
          if (m_state == LEADER) {
            m_handoverTo = CheckHandoverLeader(m_vtlList.front());
            if (m_handoverTo != VTL_INT_INVALID) {
              SetState (LEADER_HANDOVER);
              // normally the vehicle crosses the intersection after a handover
              // if the vehicle can not cross the intersection but performed a handover,
              // it must degrade when the traffic light's validity ends
              m_leaderHandoverDegradeEvent = Simulator::Schedule
                (m_vtlList.back().validUntil - Simulator::Now (), &VtlApplication::Reset, this);
              VTL_LOG_DEBUG ("Performing handover to vehicle " << m_handoverTo);
            } else {
              VTL_LOG_DEBUG ("No vehicle to handover to found");
            }
          }
        } else {
          if (HasYellow() && AfterDBrake()) {
            VTL_LOG_DEBUG ("Got yellow light after DBrake -> Crossing");
          } else {
            VTL_LOG_DEBUG ("Red light -> Stopping");
            m_mob->SetApproach (IntelligentDriverMobilityModel::RED);
            m_requiredDecelerationTrace (Simulator::Now(), GetRequiredDeceleration(m_speed, m_dist));
          }
        }
      } else {
        // vtl is not already valid but will be vaild in the future
        Simulator::Schedule(m_vtlList.front().validFrom - Simulator::Now (), &VtlApplication::CheckCurrentVtl, this);
      }
    } else {
      // if vtl is already invalid, it should have been removed by VtlExpire
      NS_ASSERT (false);
    }
  }
}


inline  bool
VtlApplication::HasGreen (void) 
{
  return (m_vtlList.front().lightmap[m_localId] == 1);
}

inline  bool
VtlApplication::HasYellow (void) 
{
  return (m_vtlList.front().lightmap[m_localId] == 2);
}


/*
 * some objective function for calculating the leader rank
 * smaller rank = better leader
 */
double
VtlApplication::GetLeaderRank (double dist, uint8_t csize)
{
  return dist - csize;
}

bool
VtlApplication::CheckBecomeLeader (void)
{
  switch (e_leaderStrat) {
    case DISTANCE:
      return CheckBecomeLeaderByDistance ();
    case DISTANCESAFE:
      return CheckBecomeLeaderByDistanceSafe ();
    case ID:
      return CheckBecomeLeaderById ();
    case PROBABILISTIC:
      return CheckBecomeLeaderProbabilistic ();
  }
  return false;
}

/*
 * requires current information on all other lanes
 * smallest ID wins
 */
bool
VtlApplication::CheckBecomeLeaderById (void)
{
  uint32_t i;
  
  for (i = 0; i < VTL_HEADER_MAXROADS; ++i) {
    // leave out own road
    if (i != m_localId) {
      // check if entry is still valid and vehicle is not already leader
      if (m_clmap[i].id != VTL_INT_INVALID && Simulator::Now () < m_clmap[i].clValidity
        && m_clmap[i].isLeader == false) {
        if (m_clmap[i].id < m_id) {
          return false; // other vehicle is better suited
        }
      } else {
        return false; // no information on road
      }
    }
  }
  return true;  // got information on all roads and I'm best leader
}

/*
 * requires current information on all other lanes
 * smallest m_dist wins
 */
bool
VtlApplication::CheckBecomeLeaderByDistanceSafe (void)
{
  uint32_t i;
  
  for (i = 0; i < VTL_HEADER_MAXROADS; ++i) {
    // leave out own road
    if (i != m_localId) {
      // check if entry is still valid and vehicle is not already leader
      if (m_clmap[i].id != VTL_INT_INVALID && Simulator::Now () < m_clmap[i].clValidity
          && m_clmap[i].isLeader == false) {
        
        double timediff = (Simulator::Now() - m_clmap[i].timestamp).GetSeconds();
        // calculate the minimum possible distance for the other vehicle, it can't be closer than that!
        double minDist = m_clmap[i].dist - m_clmap[i].speed * timediff - 0.5*m_maxAcc*timediff*timediff;
        if (minDist < m_dist) {
          return false; // other vehicle is better suited
        }
      } else {
        return false; // no information on road
      }
    }
  }
  return true;  // got information on all roads and I'm best leader
}

/*
 * requires current information on all other lanes
 * smallest m_dist wins
 */
bool
VtlApplication::CheckBecomeLeaderByDistance (void)
{
  uint32_t i;
  
  for (i = 0; i < VTL_HEADER_MAXROADS; ++i) {
    // leave out own road
    if (i != m_localId) {
      // check if entry is still valid and vehicle is not already leader
      if (m_clmap[i].id != VTL_INT_INVALID && Simulator::Now () < m_clmap[i].clValidity
          && m_clmap[i].isLeader == false) {
        
        double timediff = (Simulator::Now() - m_clmap[i].timestamp).GetSeconds();
        // calculate the expected distance for the other vehicle, it can be closer than that!
        double minDist = m_clmap[i].dist - m_clmap[i].speed * timediff;
        if (minDist < m_dist) {
          return false; // other vehicle is better suited
        }
      } else {
        return false; // no information on road
      }
    }
  }
  return true;  // got information on all roads and I'm best leader
}

/*
 * does not require information on all other lanes but only the guarantee
 * that with a certain probability no other vehicle can be on that road.
 * smallest m_dist wins
 */
bool
VtlApplication::CheckBecomeLeaderProbabilistic (void)
{
  uint32_t i;
  bool assumptions[VTL_HEADER_MAXROADS];
  
  for (i = 0; i < VTL_HEADER_MAXROADS; ++i) {
    assumptions[i] = false;
    // leave out own road
    if (i != m_localId) {
      // check if entry is still valid and vehicle is not already leader
      if (m_clmap[i].id != VTL_INT_INVALID && Simulator::Now () < m_clmap[i].clValidity
          && m_clmap[i].isLeader == false) {
        
        double timediff = (Simulator::Now() - m_clmap[i].timestamp).GetSeconds();
        // calculate the expected possible distance for the other vehicle, it can be closer than that!
        double minDist = m_clmap[i].dist - m_clmap[i].speed * timediff;
        if (minDist < m_dist) {
          return false; // other vehicle is better suited
        }
      } else {
        if (m_dist > 55)
          return false;
        Road::roadrelation o = m_navi->GetRelationTo(i);
        if (m_vtlcalc->IsRoadEmpty(m_dist, (Simulator::Now ()-m_clObservationStart).GetSeconds(), o) == false) {
          return false; // no information on road
        } else {
          assumptions[i] = true;
        }
      }
    }
  }
  
  for (i = 0; i < VTL_HEADER_MAXROADS; ++i) {
    if (assumptions[i])
      VTL_LOG_DEBUG ("CheckBecomeLeaderProbabilistic: assuming no vehicle on road localId = " << (int)i);
  }
  return true;  // got information on all roads and I'm best leader
}




uint32_t
VtlApplication::CheckHandoverLeader (const VtlInfo &vi)
{
  uint32_t i;
  
  // TODO implement other decisions here
  for (i = 0; i < VTL_HEADER_MAXROADS; ++i) {
    // leave out own road
    if (i != m_localId) {
      // check if entry is still valid
      if (m_clmap[i].id != VTL_INT_INVALID && Simulator::Now () < m_clmap[i].clValidity) {
        // check if other CL now has red light
        if (vi.lightmap[m_clmap[i].localId] == 0) {
          return m_clmap[i].id; // found one CL that has red
        }
      }
    }
  }
  return VTL_INT_INVALID;  // no vehicle to handover to found
}


bool
VtlApplication::RequireAllYellowInterval (void)
{
  uint32_t i;
  uint8_t localId;
  std::list<Ptr<Road> >::iterator rit;
  
  for (i = 0; i < m_roadmap.size(); ++i) {
    if (i != m_currentPhase) {  // ignore current phase as it will get green light
      for (rit = m_roadmap[i].begin(); rit != m_roadmap[i].end(); ++rit) {
        // this road will get red. check if there is any vehicle within dbrake
        // on that road
        localId = (*rit)->m_localId;
        if (m_clmap[localId].id != VTL_INT_INVALID 
              && Simulator::Now () < m_clmap[localId].clValidity) {
          
          double timediff = (Simulator::Now() - m_clmap[localId].timestamp).GetSeconds();
          // calculate the expected distance for the other vehicle
          double minDist = m_clmap[localId].dist - m_clmap[localId].speed * timediff;
          if (minDist < 20.0 || minDist < DSafe(m_clmap[localId].speed)) {
            return true;
          }
        }
      }
    }
  }
  
  return false;
}

uint8_t
VtlApplication::GetFirstEtaRoad (void)
{
  uint32_t i;
  double mineta;
  uint8_t minetaI;
  double eta;
  double timediff;
  
  // own eta
  eta = m_dist / m_speed;
  mineta = eta;
  minetaI = m_localId;

  for (i = 0; i < VTL_HEADER_MAXROADS; ++i) {
    // leave out own road
    if (i != m_localId) {
      // check if entry is still valid and vehicle is not already leader
      if (m_clmap[i].id != VTL_INT_INVALID && Simulator::Now () < m_clmap[i].clValidity
          && m_clmap[i].isLeader == false) {
        
        timediff = (Simulator::Now() - m_clmap[i].timestamp).GetSeconds();
        // calculate the expected time of arrival (eta) for other vehicle
        eta = m_clmap[i].dist / m_clmap[i].speed + m_clmap[i].timestamp.GetSeconds();
        if (eta < mineta) {
          mineta = eta;
          minetaI = i;
        }
        
      } else {
        // no information, no eta
      }
    }
  }

  return minetaI;
}


uint32_t
VtlApplication::FindRoadInRoadMapByLocalId(uint8_t localId)
{
  std::list<Ptr<Road> >::iterator rit;
  uint32_t i;
  for (i = 0; i < m_roadmap.size(); ++i) {
    for (rit = m_roadmap[i].begin(); rit != m_roadmap[i].end(); ++rit) {
      if ((*rit)->m_localId == localId)
        return i;
    }
  }
  NS_ASSERT(false);
  return 0;
}

/*
 * sets red light to own road and sets next change to now + given phase length
 */
void
VtlApplication::InitVtl ()
{
  NS_LOG_FUNCTION_NOARGS ();
  
  m_vtlList.clear();
  m_roadmap = m_navi->m_curInt->GetRoadMap ();
  NS_ASSERT(m_roadmap.size () > 0);
  

  
  // get the id of the road with the vehicle that will reach the intersection first
  uint8_t firstRoadId = GetFirstEtaRoad ();
  
  // locate in which TL-set this road is
  m_currentPhase = FindRoadInRoadMapByLocalId(firstRoadId);
  
  // check if we need an all-yellow interval
  if (RequireAllYellowInterval()) {
    VTL_LOG_DEBUG ("InitVtl: required AllYellow interval");
    // start with a 5 seconds all yellow interval to let failsafe passing vehicles 
    // leave intersection
    VtlInfo vi;
    vi.validFrom = Simulator::Now ();
    vi.validUntil = Simulator::Now () + Seconds(5);  // TODO make config
    memset(&vi.lightmap, 2, sizeof(vi.lightmap));
    m_vtlList.push_back(vi);
    NS_ASSERT(m_vtlList.size() <= VTL_HEADER_MAXENTRIES);
    Simulator::Schedule(vi.validUntil - Simulator::Now (), &VtlApplication::VtlExpire, this);
    
  }
  
  
  m_currentPhase--;
  LeaderUpdateTimeslice ();
  
  // set approach state
  CheckCurrentVtl();
  
}

void
VtlApplication::LeaderUpdateTimeslice()
{
  NS_ASSERT (m_leaderUpdateEvent.IsExpired ());
  
  if (m_state == FINISHED)
    return;
    
  Time validFrom;
  std::list<Ptr<Road> >::iterator rit;
  
  m_currentPhase++;
  if (m_currentPhase == m_roadmap.size ())
    m_currentPhase = 0;
  
  
  if (m_vtlList.empty()) {
    validFrom = Simulator::Now ();
  } else {
    validFrom = m_vtlList.back().validUntil;
  }
  VtlInfo vi;
  
  vi.validFrom = validFrom;
  
  // estimate optimum phase length
  uint8_t maxClusterSize = 0;
  uint8_t clusterSize;
  for (rit = m_roadmap[m_currentPhase].begin(); rit != m_roadmap[m_currentPhase].end(); ++rit) {
    if ((*rit)->m_localId == m_localId) {
      clusterSize = m_ownCluster.size() + 1;
    } else {
      clusterSize = m_clmap[(*rit)->m_localId].clusterSize + 1;
    }
    if (clusterSize > maxClusterSize)
      maxClusterSize = clusterSize;
  }
  Time phaselength = Seconds(2 + maxClusterSize * 3);
  
  if (phaselength > m_maxPhaseLength)
    phaselength = m_maxPhaseLength;
  if (phaselength < m_minPhaseLength)
    phaselength = m_minPhaseLength;
      
  VTL_LOG_DEBUG ("LeaderUpdateTimeslice: maxClusterSize = " << (int)maxClusterSize 
        << " phaselength = " << phaselength.GetSeconds());
  
  vi.validUntil = validFrom + phaselength;
  memset(&vi.lightmap, 0, sizeof(vi.lightmap));
  
  for (rit = m_roadmap[m_currentPhase].begin(); rit != m_roadmap[m_currentPhase].end(); ++rit) {
    vi.lightmap[(*rit)->m_localId] = 1;
    VTL_LOG_DEBUG ("LeaderUpdateTimeslice: green on roadId = " << (*rit)->m_id);
  }
  m_vtlList.push_back(vi);
  NS_ASSERT(m_vtlList.size() <= VTL_HEADER_MAXENTRIES);
  Simulator::Schedule(vi.validUntil - Simulator::Now (), &VtlApplication::VtlExpire, this);
  
  
  
  // TODO make this config
  VTL_LOG_DEBUG("schedule next LeaderUpdateTimeslice in " << (vi.validUntil - m_leaderTimeslicePretime - Simulator::Now ()).GetSeconds() << " Seconds");
  m_leaderUpdateEvent = Simulator::Schedule(vi.validUntil - m_leaderTimeslicePretime - Simulator::Now (), &VtlApplication::LeaderUpdateTimeslice, this);
}


void
VtlApplication::UpdatePositionAndSpeed ()
{
  NS_LOG_FUNCTION_NOARGS ();
  m_pos = m_mob->GetPosition ();
  m_speed = m_mob->GetSpeed ();
  m_dist = m_navi->GetStopDist ();
  m_acc = m_mob->GetAcc ();
}

inline bool
VtlApplication::AfterDBrake()
{
  double timediff = m_interval.GetSeconds();
  return DSafe(m_speed+m_maxAcc*timediff) > m_dist - m_speed * timediff 
    - 0.5*m_maxAcc*timediff*timediff;
}

inline double
VtlApplication::DSafe(double speed)
{
  return ((speed*speed)/(2.0*m_safeAcc) + speed * m_reactionTime);
}

inline double
VtlApplication::GetRequiredDeceleration(double speed, double distance)
{
  if (distance > 0)
    return speed*speed / (2.0*distance);
  return 0;
}

/*
 * called every timestep
 * updates own position and speed
 * updates vtl status
 * broadcasts beacon
 */
void 
VtlApplication::NextTimestep ()
{
  NS_LOG_FUNCTION_NOARGS ();
  
  if (m_state != FINISHED || m_navi->m_running == true) {
    UpdatePositionAndSpeed ();
  }
  

  // Intersection passing (yellow handling)
  // check if make it across intersection at dsafe
  if (m_approach == VALIDVTL) {
    // check if we can make it across intersection at D_safe
    if (HasGreen () && AfterDBrake ()) {
      double timeleft = (m_vtlList.front().validUntil - Simulator::Now()).GetSeconds();
      if (timeleft > VTL_MIN_SECONDS_DBRAKE ||
            (m_speed * timeleft + (0.5 * m_acc * timeleft * timeleft)
            >= m_navi->GetCompleteCrossingDist ())) {
        // we can make it!
        m_approach = VALIDVTLDECIDED;
        VTL_LOG_DEBUG("At DBrake: GO!" 
        << " TimeLeft = " << timeleft
        << " Speed = " << m_speed
        << " Acceleration = " << m_acc
        << " CompleteCrossingDistance = " << m_navi->GetCompleteCrossingDist ());
      } else {
        m_mob->SetApproach (IntelligentDriverMobilityModel::RED);
        m_requiredDecelerationTrace (Simulator::Now(), GetRequiredDeceleration(m_speed, m_dist));
        m_approach = VALIDVTLDECIDED;
        VTL_LOG_DEBUG("At DBrake: STOP!" 
        << " TimeLeft = " << timeleft
        << " Speed = " << m_speed
        << " Acceleration = " << m_acc
        << " CompleteCrossingDistance = " << m_navi->GetCompleteCrossingDist ());
      }
    }
  }


  if (!m_disabled) {
    // CLUSTERING
    // check if we should become CL
    if (m_state == SLAVE && (Simulator::Now () - m_foundBetterClusterLeaderTime) > m_clusterInterval) {
      m_clTrace (Simulator::Now());
      SetState (CL);
      // reset all information on other cluster leaders.
      // this is the equivalent to promela election ready flag under the assumption of no delay
      ResetOtherCL ();
    }
    
    
    // Check if we can become VTL Leader
    if (m_approach == FAILSAFE && m_state == CL) {
      if (CheckBecomeLeader ()) {
        m_vtllElectionTrace (Simulator::Now());
        
        SetState (LEADER);
        InitVtl ();
      }
    }

    double jitterInSeconds = jitterGenerator.GetValue();
    Time jitter = Seconds(jitterInSeconds);
    m_sendEvent = Simulator::Schedule(jitter, &VtlApplication::SendPacket, this);

  } // (!m_disabled)

  if (m_state == FINISHED && m_navi->m_running == false) {
    m_finishedTrace(Simulator::Now(), (Simulator::Now() - m_startTime).GetSeconds());
    VTL_LOG_DEBUG ("Vehicle finished!");
    StopApplication();
  } else {
    ScheduleNextTx();
    ApproachState a = (m_approach == FAILSAFE) ? FAILSAFE : (HasGreen() ? GREEN : RED);
    m_vtlTrace(Simulator::Now (), m_id, m_pos, m_dir, m_speed, m_state, a);
  } 
  

}



Time
VtlApplication::calcClValidity(void)
{
  double ts;
  double minInterval = 0.2;
  ts = (std::sqrt(m_speed*m_speed + 2*m_dist*m_maxAcc) - m_speed) / m_maxAcc;
  if (ts < minInterval) {
    return Simulator::Now () + Seconds(minInterval);
  } else {
    return Simulator::Now () + Seconds(ts);
  }
}

void
VtlApplication::SendPacket()
{
  NS_LOG_FUNCTION_NOARGS ();
//  NS_LOG_INFO ("VtlApplication::SendPacket() at " << Simulator::Now().GetSeconds() << "sec");
  NS_ASSERT (m_sendEvent.IsExpired ());
  Ptr<Packet> packet;

  packet = Create<Packet> (m_packetSize);;

  
    // Attach VTL header to packet
  VtlHeader vtlHdr;
  vtlHdr.m_source = m_id;
  vtlHdr.m_seqNo = m_seqNo++;
  vtlHdr.m_timestamp = Simulator::Now();
  vtlHdr.m_pos = m_pos;
  vtlHdr.m_speed = m_speed;
  vtlHdr.m_intersectionId = m_intersectionId;
  vtlHdr.m_localId = m_localId;
  if (m_state == CL || m_state == LEADER || m_state == LEADER_HANDOVER) {
    vtlHdr.m_isClusterLeader = true;
    vtlHdr.m_clusterSize = m_ownCluster.size();
  } else {  // SLAVE or FINISHED
    vtlHdr.m_isClusterLeader = false;
  }
  vtlHdr.m_clValidity = calcClValidity ();

  vtlHdr.m_isLeader = (m_state == LEADER || m_state == LEADER_HANDOVER);
  vtlHdr.m_dist = m_dist;
  vtlHdr.m_nextLeader = m_handoverTo;
  vtlHdr.m_currentPhase = m_currentPhase;



  uint32_t rebroadcaststrategy = 1;
  bool rebroadcast;
  switch (rebroadcaststrategy) {
    case 0: // Off (Leader only)
      rebroadcast = (m_state == LEADER || m_state == LEADER_HANDOVER);
      break;
    default:
    case 1: // Leader + CL
      rebroadcast = (m_state == CL || m_state == LEADER || m_state == LEADER_HANDOVER);
      break;
    case 2: // All
      rebroadcast = true;
      break;
    case 3: // Some fancy policy
      rebroadcast = true;
      break;
  }
    

  if (rebroadcast) {
    uint32_t i = 0;
    std::list<VtlInfo>::iterator it;
    for (it = m_vtlList.begin(); it != m_vtlList.end(); ++it) {
      vtlHdr.m_vtlInfo[i++] = *it;
      //VTL_LOG_DEBUG ("added VtlInfo valid from " << it->validFrom.GetSeconds() << " valid until " << it->validUntil.GetSeconds() );
    }
    vtlHdr.m_nVtlInfo = i;
  } else {
    vtlHdr.m_nVtlInfo = 0;
  }
  
  
  packet->AddHeader(vtlHdr);


  int rc;
  rc = m_socket->Send (packet);
  
  if (rc < 0) {
    VTL_LOG_DEBUG ( "###################socket error" << (Socket::SocketErrno) m_socket->GetErrno());
  }
  

}


} // Namespace ns3
