// number of processes
#define NPROC 3   // CHANGE NPROCROADS

//number of roads
#define NROADS 2   // CHANGE NPROCROADS

// NROADS * NPROC, must not be done manually anymore!
#define NPROCROADS NPROC*NROADS


mtype = {BEACON, VTLINFO, HANDOVER}

// we need one channel per process
chan chanA = [1] of { mtype, byte, byte, bool, byte, byte };
#if NPROC >= 2
chan chanB = [1] of { mtype, byte, byte, bool, byte, byte };
#endif
#if NPROC >= 3
chan chanC = [1] of { mtype, byte, byte, bool, byte, byte };
#endif
#if NPROC >= 4
chan chanD = [1] of { mtype, byte, byte, bool, byte, byte };
#endif
#if NPROC >= 5
chan chanE = [1] of { mtype, byte, byte, bool, byte, byte };
#endif
#if NPROC >= 6
chan chanF = [1] of { mtype, byte, byte, bool, byte, byte };
#endif


// number of initialized vehicles
byte proccount=0;

// Ensure that only one vehicle per road is CL (i.e., first vehicle)
// by letting only one vehicle per road set this to its ID
// When the CL leaves the intersection, it sets the value to 255 and other
// vehicles can regularly check if they become CL.
byte clusterLeader[NROADS];

// vehicles store the id of other CLs in this 2-dimensional array
// when a vehicle leaves the intersection, this information must be set to 255
byte informationMap[NPROCROADS];


// each vehicle stores the id of the currently green road in this array
// When the validity of the VTL expires, each value of this array will
// be set to 255.
byte greenRoad[NPROC];


// set to true if vehicle is VTL Leader
// used to find out, if beacon messages are outdated, i.e., sent before vehicle
// became VTL Leader but received after
// in reality, limited validity of beacons could do this and a vehicle must not
// become VTL Leader until its beacons are invalid
bool isLeader[NPROC];


inline broadcast(tmsg,tid,troad,tcl,tgreen,thandover)
{
  atomic {  // require indeterminism for packetloss
  
  skip; // DONT REMOVE, spin says unexecutable otherwise
  printf("vehicle %u broadcast\n", id+1);
      
#if NPROC >= 1
  // send packet from vehicle with ID 'id' to vehicle 0, i.e., chanA
  //assert(0);
  if
  :: (id == 0)  // do not send to own chan
  :: else ->
    if
    :: skip  // MAY drop packet if between roads
    :: skip -> chanA!tmsg,tid,troad,tcl,tgreen,thandover  // MAY send to everyone else
    fi
  fi;
#endif

#if NPROC >= 2
  if
  :: (id == 1)
  :: else ->
    if
    :: skip  // MAY drop packet if between roads
    :: skip -> chanB!tmsg,tid,troad,tcl,tgreen,thandover  // MAY send to everyone else
    fi
  fi; 
#endif

#if NPROC >= 3
  if
  :: (id == 2)
  :: else ->
    if
    :: skip  // MAY drop packet if between roads
    :: skip -> chanC!tmsg,tid,troad,tcl,tgreen,thandover  // MAY send to everyone else
    fi
  fi;
#endif

#if NPROC >= 4
  if
  :: (id == 3)
  :: else ->
    if
    :: skip  // MAY drop packet if between roads
    :: skip -> chanD!tmsg,tid,troad,tcl,tgreen,thandover  // MAY send to everyone else
    fi
  fi
#endif

#if NPROC >= 5
  if
  :: (id == 4)
  :: else ->
    if
    :: skip  // MAY drop packet if between roads
    :: skip -> chanE!tmsg,tid,troad,tcl,tgreen,thandover  // MAY send to everyone else
    fi
  fi
#endif

#if NPROC >= 6
  if
  :: (id == 5)
  :: else ->
    if
    :: skip  // MAY drop packet if between roads
    :: skip -> chanF!tmsg,tid,troad,tcl,tgreen,thandover  // MAY send to everyone else
    fi
  fi
#endif
  }
}

// checks main assertion whenever VTL info is changed somewhere:
// There exists one VTL info so that for all vehicles V: 
// V has that VTL info or V has no info
inline checkAssert()
{
  d_step {
    i=0;
    tgreen=255;
    do		
    :: i >= NPROC -> break
		:: else -> 
		  if
		  :: (tgreen == 255 && greenRoad[i] != 255) ->
		      tgreen = greenRoad[i]
		  :: else
		  fi;
		  assert(greenRoad[i] == tgreen || greenRoad[i] == 255);
		  i++
		od;
		i=0;
		tgreen=255
	}
}

// only one vehicle per road is CL, this is invoked in during the send/receive
// cycle
inline checkCL()
{
    if
    :: (clusterLeader[road] == 255) -> 
      clusterLeader[road] = id; 
      isCL = 1;
      printf("vehicle %u became CL\n", id+1)
    :: else
    fi
}

// reset information that other vehicles have about me
// assume: vehicles do not have outdated information
inline deleteCLinfo()
{
  d_step {
    printf("vehicle %u information delete\n", id+1);
    i = 0;
    do
    :: i >= NPROC -> break
    :: else -> informationMap[i*NROADS+road] = 255; i++
    od;
    i = 0;
  }
}

// main send receive cycle when approaching intersection
// uses d_step so don't nest within atomic {...}
inline sendrecv()
{
  do
  :: d_step {
      checkCL();
    }
  
  :: // send
    atomic {
      broadcast(BEACON,id,road,isCL,0,0);
    }
  
  
  // receive
  :: atomic {
    myChan?msg,senderId,senderRoad,senderIsCL,senderGreenRoadId,senderHandoverTo;
    // store VTL information
    if
    :: (msg == VTLINFO) -> 
        if
        :: greenRoad[senderId] == senderGreenRoadId ->
            greenRoad[id] = senderGreenRoadId;
            checkAssert()
        :: else // VTL info already expired
        fi
        
 
      // two validity constraints must be met for a handover:
      // 1. the handover information must still be valid, e.g. the old leader
      // must not have finished already (check if sender is still CL)
      // 2. the CL information at other vehicles must be invalid before the new
      // vehicles becomes leader (atomic deleteCLinfo & set isLeader)
      // in reality this means: become leader AFTER own beacons expire but
      // BEFORE handover expired
        
    :: (msg == HANDOVER && senderHandoverTo == id && clusterLeader[senderRoad] == senderId) ->
        greenRoad[id] = senderGreenRoadId;
        isLeader[id] = 1;
        deleteCLinfo();
        byhandover = true;
        goto leader
    :: else
    fi;
    }
    
    d_step {
      // store other CL information
      // we have to check if sender is still CL to account for delayed packets
      // and if sender did not become VTL Leader in the meantime
      // must be done with validity in reality
      if
      :: (senderIsCL == 1 && clusterLeader[senderRoad] == senderId && isLeader[senderId] == 0) ->
         informationMap[id*NROADS+senderRoad] = senderId
      :: else
      fi;
    
    printf("%u received %e from %u\n", id+1, msg, senderId+1)
    }
    
  :: break
  od;
}

inline invalidateVTLinfo()
{
  // invalidate vtl information
  d_step {
    i=0;
    do		
    :: i >= NPROC -> break
    :: else -> 
      // when invalidating others information, ensure that they have
      // either my information or no information (but not some other information)
      assert(greenRoad[i] == lgreen || greenRoad[i] == 255);
      greenRoad[i] = 255;
      i++
    od;
    i=0
  }
}

// assigns green light to next road and invalidates old VTL information
// which equals 'validity expired' event in real protocol
inline nextGreen()
{
  d_step {
    printf("Vehicle %u at Nextgreen\n", id+1);

    invalidateVTLinfo();

		// give green to next road
    lgreen = lgreen + 1;
    if
    :: lgreen >= NROADS ->
      lgreen = 0;
    :: else
    fi;
    greenRoad[id] = lgreen;
    checkAssert()

  }
}


proctype vehicle(byte id)
{
  // state variables
  byte road;
  bool isCL = 0;
  chan myChan;
  bool byhandover = 0;

  // message variables
  mtype msg;
  byte senderId;
  byte senderRoad;
  bool senderIsCL;
  byte senderGreenRoadId;
  byte senderHandoverTo;
  
  // VTL leader related
  byte lgreen;
  byte handoverto;
  
  // misc
  byte i;
  byte maxId;
  byte tgreen;
      
  // the vehicle can be on either one of the conflicting roads
  atomic {
    if 
    :: (NROADS >=1) -> road = 0
    :: (NROADS >=2) -> road = 1
    :: (NROADS >=3) -> road = 2
    :: (NROADS >=4) -> road = 3
    fi;
  }
  d_step {

    if
    :: id == 0 -> myChan = chanA; //xr chanA
#if NPROC >= 2
    :: id == 1 -> myChan = chanB; //xr chanB
#endif
#if NPROC >= 3
    :: id == 2 -> myChan = chanC; //xr chanC
#endif
#if NPROC >= 4
    :: id == 3 -> myChan = chanD; //xr chanD
#endif
#if NPROC >= 5
    :: id == 4 -> myChan = chanE; 
#endif
#if NPROC >= 6
    :: id == 5 -> myChan = chanF; 
#endif
    fi;
    
    // wait for all processes to be initialized
    proccount++;
  }
    (proccount == NPROC);
  
  
  // main send/receive cycle when vehicle is approaching intersection
  sendrecv();

  
  
  // vehicle is at D_Brake
dbrake:
  atomic {
    printf("Vehicle %u at D_Brake\n", id+1);
    if
    // vehicle has valid VTL information
    :: (greenRoad[id] != 255) ->
      if
      // vehicle has green light
      :: (greenRoad[id] == road) ->
        goto green
      // vehicle has red light
      :: (greenRoad[id] != road) ->
        goto red
      fi;
    
    
    // no valid VTL information
    :: (greenRoad[id] == 255) ->
      if
      // vehicle is first on road
      :: (isCL == 1) -> // check if information on all other roads is available
        goto checkLeader
        
      // vehicle is not CL
      :: (isCL == 0) ->
        goto failsafe
      fi;
    fi;
  }

// should not reach here


////////////////
// CHECKLEADER
////////////////
checkLeader:
  skip;
  atomic {
    d_step {
      i=0;
      maxId=0;
      do
        :: (i >= NROADS) -> break;
        :: else ->
          if
            :: (i != road && informationMap[id*NROADS+i] > maxId) ->
              maxId = informationMap[id*NROADS+i]
            :: else
          fi;
          i++
      od;
      i=0;
    }
    
    if
    :: (maxId < id) ->  //become VTL Leader if I have highest ID
      isLeader[id] = 1;
      deleteCLinfo(); // see 05/06
      goto leader
    :: else ->
      goto failsafe
    fi
  }
  
////////////////
// RED
//////////////// 
red:
  printf("Vehicle %u at Red\n", id+1);
  sendrecv();
  atomic{
    if
    :: (greenRoad[id] == road) -> // i have green light now
      goto green
    :: (greenRoad[id] == 255) ->  // VTL expired
      goto failsafe
    :: else ->  // still red
      goto red
    fi;
  }
  
////////////////
// GREEN
//////////////// 
green:
  printf("Vehicle %u at Green\n", id+1);
  do
  
  ::atomic {  // send as long as we have green light (not so important)
      (greenRoad[id] == road);
      broadcast(BEACON,id,road,isCL,0,0);
      printf("%u sent %e\n", id, msg);
    }
  
  :: break  // leave intersection
  od;
  
  goto finish;


////////////////
// FAILSAFE
//////////////// 
failsafe:
  printf("Vehicle %u at Failsafe\n", id+1);
  sendrecv();
  atomic {
    if
      :: goto finish
      :: goto dbrake
    fi;
  }
////////////////
// LEADER
//////////////// 
leader:
  skip;
  d_step {
  printf("Vehicle %u at Leader\n", id+1);
  
  // see 03/15 problem handover if this does not happen
  // this is now done BEFORE the goto leader because of atomicity!
  //isLeader[id] = 1;
  //deleteCLinfo();
  
  // initialize VTL by giving green to other road
    if
    :: (byhandover == 1 && greenRoad[id] != 255) ->
        skip  // old VTL is still valid so wait until is has become
              // invalid before changing it
    :: else
        lgreen = road;
        nextGreen()
    fi

  }

  
  do
  // send VTL information
  :: broadcast(VTLINFO,id,road,isCL,lgreen,0)
  
  // next road gets green
  // the guard prevents changing the vtl too early
  // after a handover, i.e., when the old leader is still
  // in charge
  :: (byhandover == 0 || greenRoad[id] == 255) ->
    byhandover = 0;
    nextGreen();
    if
    :: (lgreen == road) ->
      goto handover 
      //goto finish
    :: (lgreen == road) // added 16.4. to reduce number of unreached states
    :: else
    fi

    
  // receive others information
  :: myChan?msg,senderId,senderRoad,senderIsCL,senderGreenRoadId,senderHandoverTo;
    d_step {
      // store other CL information
      // check if CL is still CL, ie validity not yet expired
      if
      :: (senderIsCL == 1 && clusterLeader[senderRoad] == senderId) ->
         informationMap[id*NROADS+senderRoad] = senderId
      :: else
      fi;
      
    }
  od;

////////////////
// HANDOVER
//////////////// 
handover:
  // check if other CL is available
  skip;
  d_step {
    printf("Vehicle %u at Handover\n", id+1);
    i=0;
    handoverto = 255;
    do
    :: (i >= NROADS) -> break
    :: else ->
      if
        :: (i != road && informationMap[id*NROADS+i] != 255) ->
          handoverto = informationMap[id*NROADS+i];
          break
        :: else
      fi;
      i++
    od;
    i=0;
  }
  


  // broadcast VTL info and (maybe) handover message
  do
  
  :: broadcast(VTLINFO,id,road,isCL,lgreen,0)
  
  :: (handoverto != 255) ->  // send handover if other vehicle is there
    broadcast(HANDOVER,id,road,isCL,lgreen,handoverto)

    
  ::atomic {  // receive
    myChan?msg,senderId,senderRoad,senderIsCL,senderGreenRoadId,senderHandoverTo;
    // store VTL information
    if
    :: (msg == VTLINFO) -> 
        if
        :: greenRoad[senderId] == senderGreenRoadId ->
            greenRoad[id] = senderGreenRoadId;
            checkAssert()
        :: else // VTL info already expired
        fi;
        //goto dbrake
    :: else
    fi;
    }
    
  :: break
  od;


  // my VTL expired, next Leader is in charge
  invalidateVTLinfo();


////////////////
// FINISH
//////////////// 
finish:
  atomic {
    printf("Vehicle %u at Finish\n", id+1);
    if
    :: (isCL == 1) ->
      // next vehicle can now become CL
      clusterLeader[road] = 255;
      isLeader[id] = 0;
      deleteCLinfo()
    :: else
    fi
  }
}


init
{
  int i;

  d_step {
    assert (NPROC <= 6);  // need to change for more procs
  
    i=0;
    do		
    :: i >= NPROC -> break
		:: else -> greenRoad[i] = 255; i++
		od;		
		
    i=0;
		do		
    :: i >= NPROC -> break
		:: else -> isLeader[i] = 0; i++
		od;
		
    i=0;
    do		
    :: i >= NROADS -> break
		:: else -> clusterLeader[i] = 255; i++
		od;
		
    i=0;
    do		
    :: i >= NROADS*NPROC -> break
		:: else -> informationMap[i] = 255; i++
		od;  
	}
	
	atomic {
    i=0;
    do
		:: i >= NPROC -> break
		:: else -> run vehicle(i); i++
	  od;
	  
  }
}
